import os

import aiosqlite

from assistants.config.file_management import DB_PATH
from assistants.log import logger
from assistants.user_data.sqlite_backend.assistants import TABLE_NAME as ASSISTANTS
from assistants.user_data.sqlite_backend.chat_history import TABLE_NAME as CHAT_HISTORY
from assistants.user_data.sqlite_backend.conversations import conversations_table
from assistants.user_data.sqlite_backend.telegram_chat_data import telegram_data
from assistants.user_data.sqlite_backend.threads import TABLE_NAME as THREADS


async def table_exists(db_path, table_name):
    async with aiosqlite.connect(db_path) as db:
        try:
            # Query to check for the table in sqlite_master
            async with db.execute(
                "SELECT name FROM sqlite_master WHERE type='table' AND name=?;",
                (table_name,),
            ) as cursor:
                result = await cursor.fetchone()
                return result is not None  # If result is not None, the table exists
        except aiosqlite.Error as e:
            print(f"An error occurred while checking for the table: {e}")
            return False


async def column_exists(db_path, table_name, column_name):
    async with aiosqlite.connect(db_path) as db:
        try:
            # Execute the PRAGMA statement to get column info
            async with db.execute(f"PRAGMA table_info({table_name})") as cursor:
                columns = await cursor.fetchall()
                # Check if the column exists in the returned columns
                for column in columns:
                    if column[1] == column_name:  # column[1] is the column name
                        return True
        except aiosqlite.Error as e:
            print(f"An error occurred: {e}")
            # Return False or handle the error as needed, e.g., table doesn't exist.
            return False
    return False


async def drop_table(db_path, table_name):
    async with aiosqlite.connect(db_path) as db:
        try:
            # Execute the DROP TABLE statement
            await db.execute(f"DROP TABLE IF EXISTS {table_name};")
            await db.commit()  # Commit the changes
            print(f"Table '{table_name}' has been dropped successfully.")
        except aiosqlite.Error as e:
            print(f"An error occurred while dropping the table: {e}")


async def init_db():
    if not DB_PATH.parent.exists():
        DB_PATH.parent.mkdir(parents=True)

    async with aiosqlite.connect(DB_PATH) as db:
        await db.execute(
            f"CREATE TABLE IF NOT EXISTS {ASSISTANTS} (assistant_name TEXT PRIMARY KEY, assistant_id TEXT, config_hash TEXT);"
        )
        await db.execute(
            f"CREATE TABLE IF NOT EXISTS {THREADS} (thread_id TEXT PRIMARY KEY, assistant_id TEXT, last_run_dt TEXT, initial_prompt TEXT);"
        )

        await db.commit()

        await conversations_table.create_table()

        if os.getenv("TELEGRAM_DATA"):
            # Handle an old version of DB where this table was created in error
            if await table_exists(DB_PATH, CHAT_HISTORY):
                if not await column_exists(DB_PATH, CHAT_HISTORY, "thread_id"):
                    await drop_table(DB_PATH, CHAT_HISTORY)

            await telegram_data.create_db()


async def rebuild_db():
    if DB_PATH.exists():
        # Create backup of existing database in /tmp
        backup_file = DB_PATH.with_suffix(".bak")
        backup_file.write_bytes(DB_PATH.read_bytes())
        os.rename(backup_file, f"/tmp/{backup_file.name}")
        logger.info(f"Existing database backed up to /tmp/{backup_file.name}")
        DB_PATH.unlink()

    if DB_PATH.exists():
        raise RuntimeError("Failed to delete existing database")

    await init_db()



"""
The CLI module is the entry point for the Assistant CLI.
It is responsible for parsing command line arguments, creating the Assistant object,
and starting the IO loop.
"""

import asyncio
import select
import sys

from assistants import version
from assistants.cli import output
from assistants.cli.arg_parser import get_args
from assistants.cli.io_loop import io_loop
from assistants.cli.utils import (
    create_assistant_and_thread,
    get_text_from_default_editor,
)
from assistants.config import environment
from assistants.lib.exceptions import ConfigError


def cli():
    """
    Main function (entrypoint) for the Assistant CLI.
    """

    # Parse command line arguments, if --help is passed, it will exit here
    args = get_args()

    if select.select([sys.stdin], [], [], 0.0)[0]:
        stdin = sys.stdin.read()
    else:
        stdin = None
    if stdin:
        args.prompt = args.prompt or []
        args.prompt += stdin.split(" ")

    # Join all the positional arguments into a single string
    initial_input = " ".join(args.prompt) if args.prompt else None

    # First line of output (version and basic instructions)
    output.default(
        f"""Assistant CLI v{version.__VERSION__}; using '{environment.CODE_MODEL if args.code else environment.DEFAULT_MODEL}' model{' (reasoning model)' if args.code else ''}.
Type '/help' (or '/h') for a list of commands.
"""
    )
    output.default("")
    if args.editor:
        # Open the default editor to compose formatted prompt
        initial_input = get_text_from_default_editor(initial_input)

    elif args.input_file:
        # Read the initial prompt from a file
        try:
            with open(  # pylint: disable=unspecified-encoding
                args.input_file, "r"
            ) as file:
                initial_input = file.read()
        except FileNotFoundError:
            output.fail(f"Error: The file '{args.input_file}' was not found.")
            sys.exit(1)

    # Create assistant and get the last thread if one exists
    try:
        assistant, thread_id = asyncio.run(create_assistant_and_thread(args))
    except ConfigError as e:
        output.fail(f"Error: {e}")
        sys.exit(1)

    if thread_id is None and args.continue_thread:
        output.warn("Warning: could not read last thread id; starting new thread.")

    # IO Loop (takes user input and sends it to the assistant, or parses it as a command,
    # then prints the response before looping to do it all over again)
    try:
        io_loop(assistant, initial_input, thread_id=thread_id)
    except (EOFError, KeyboardInterrupt):
        # Exit gracefully if ctrl+C or ctrl+D are pressed
        sys.exit(0)



"""
This module defines classes for interacting with the OpenAI API(s), including memory management functionality through the MemoryMixin class.

Classes:
    - Assistant: Encapsulates interactions with the OpenAI Assistants API.
    - Completion: Encapsulates interactions with the OpenAI Chat Completion API.
"""

import asyncio
import hashlib
from typing import Optional

import openai
from openai._types import NOT_GIVEN, NotGiven
from openai.types.beta import Thread
from openai.types.beta.threads import Message, Run
from openai.types.chat import ChatCompletionMessage

from assistants.ai.memory import MemoryMixin
from assistants.ai.types import MessageData, MessageDict
from assistants.config import environment
from assistants.lib.exceptions import ConfigError, NoResponseError
from assistants.log import logger
from assistants.user_data import threads_table
from assistants.user_data.sqlite_backend.assistants import (
    get_assistant_data,
    save_assistant_id,
)
from assistants.user_data.sqlite_backend.threads import get_last_thread_for_assistant


class Assistant:  # pylint: disable=too-many-instance-attributes
    """
    Encapsulates interactions with the OpenAI Assistants API.

    Fits AssistantProtocol: Protocol defining the interface for assistant classes.

    Attributes:
        name (str): The name of the assistant.
        model (str): The model to be used by the assistant.
        instructions (str): Instructions for the assistant.
        tools (list | NotGiven): Optional tools for the assistant.
        api_key (str): API key for OpenAI.
        client (openai.OpenAI): Client for interacting with the OpenAI API.
        _config_hash (Optional[str]): Hash of the current configuration.
        assistant (Optional[object]): The assistant object.
        last_message (Optional[str]): ID of the last message in the thread.
    """

    REASONING_MODELS = ["o1", "o3-mini"]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        name: str,
        model: str,
        instructions: str,
        tools: list | NotGiven = NOT_GIVEN,
        api_key: str = environment.OPENAI_API_KEY,
        reasoning_effort: str = "medium",
    ):
        """
        Initialize the Assistant instance.

        :param name: The name of the assistant.
        :param model: The model to be used by the assistant.
        :param instructions: Instructions for the assistant.
        :param tools: Optional tools for the assistant.
        :param api_key: API key for OpenAI.
        """
        if not api_key:
            raise ConfigError("Missing 'OPENAI_API_KEY' environment variable")

        self.client = openai.OpenAI(
            api_key=api_key, default_headers={"OpenAI-Beta": "assistants=v2"}
        )
        self.instructions = instructions
        self.model = model
        self.tools = tools if self.model not in self.REASONING_MODELS else []
        self.name = name
        self._config_hash = None
        self.assistant = None
        self.last_message = None
        self.last_prompt = None
        self.reasoning_effort = (
            reasoning_effort if self.model in self.REASONING_MODELS else None
        )

    async def start(self):
        """
        Load the assistant_id from DB if exists or create a new assistant.
        """
        if not self.__dict__.get("assistant"):
            self.assistant = await self.load_or_create_assistant()
        self.last_message = None

    async def async_get_conversation_id(self):
        if self.last_message:
            return self.last_message.thread_id
        return await get_last_thread_for_assistant(self.assistant_id)

    def __getattribute__(self, item):
        """
        Override to check if the assistant is loaded before accessing it.

        :param item: The attribute name.
        :return: The attribute value.
        """
        if item == "assistant":
            if self.__dict__.get("assistant") is None:
                raise AttributeError("Assistant not loaded. Call `start` method.")
        return super().__getattribute__(item)

    @property
    def assistant_id(self):
        """
        Get the assistant ID.

        :return: The assistant ID.
        """
        return self.assistant.id

    @property
    def config_hash(self):
        """
        A hash of the current config options to prevent regeneration of the same assistant.

        :return: The configuration hash.
        """
        if not self._config_hash:
            self._config_hash = self._generate_config_hash()
        return self._config_hash

    def _generate_config_hash(self):
        """
        Generate a hash based on the current configuration.

        :return: The generated hash.
        """
        return hashlib.sha256(
            f"{self.name}{self.instructions}{self.model}{self.tools}".encode()
        ).hexdigest()

    async def load_or_create_assistant(self):
        """
        Get any existing assistant ID / config hash from the database or create a new assistant.

        :return: The assistant object.
        """
        existing_id, config_hash = await get_assistant_data(self.name, self.config_hash)
        if existing_id:
            try:
                assistant = self.client.beta.assistants.retrieve(existing_id)
                if config_hash != self.config_hash:
                    logger.info("Config has changed, updating assistant...")
                    self.client.beta.assistants.update(
                        existing_id,
                        instructions=self.instructions,
                        model=self.model,
                        tools=self.tools,
                        name=self.name,
                        reasoning_effort=self.reasoning_effort,
                    )
                    await save_assistant_id(self.name, assistant.id, self.config_hash)
                return assistant
            except openai.NotFoundError:
                pass
        assistant = self.client.beta.assistants.create(
            name=self.name,
            instructions=self.instructions,
            model=self.model,
            tools=self.tools,
            reasoning_effort=self.reasoning_effort,
        )
        await save_assistant_id(self.name, assistant.id, self.config_hash)
        return assistant

    def _create_thread(self, messages=NOT_GIVEN) -> Thread:
        """
        Create a new thread.

        :param messages: Optional initial messages for the thread.
        :return: The created thread.
        """
        return self.client.beta.threads.create(messages=messages)

    def _new_thread(self) -> Thread:
        """
        Create a new thread.

        :return: The created thread.
        """
        thread = self._create_thread()
        return thread

    def start_thread(self, prompt: str) -> Thread:
        """
        Create a new thread and add the first message to it.

        :param prompt: The initial prompt for the thread.
        :return: The created thread.
        """
        logger.debug(f"Starting new thread with prompt: {prompt}")
        thread = self._new_thread()
        self.client.beta.threads.messages.create(
            thread_id=thread.id,
            role="user",
            content=prompt,
        )
        return thread

    def continue_thread(self, prompt: str, thread_id: str) -> Message:
        """
        Add a new message to an existing thread.

        :param prompt: The message content.
        :param thread_id: The ID of the thread to continue.
        :return: The created message.
        """
        message = self.client.beta.threads.messages.create(
            thread_id=thread_id,
            role="user",
            content=prompt,
        )
        return message

    def run_thread(self, thread: Thread) -> Run:
        """
        Run the thread using the current assistant.

        :param thread: The thread to run.
        :return: The run object.
        """
        run = self.client.beta.threads.runs.create(
            thread_id=thread.id,
            assistant_id=self.assistant.id,
        )
        return run

    def _get_thread(self, thread_id: str) -> Thread:
        """
        Retrieve a thread by its ID.

        :param thread_id: The ID of the thread to retrieve.
        :return: The retrieved thread.
        """
        return self.client.beta.threads.retrieve(thread_id)

    async def prompt(self, prompt: str, thread_id: Optional[str] = None) -> Run:
        """
        Create a message using the prompt and add it to an existing thread or create a new thread.

        :param prompt: The message content.
        :param thread_id: Optional ID of the thread to continue.
        :return: The run object.
        """
        self.last_prompt = prompt
        if thread_id is None:
            thread = self.start_thread(prompt)
            run = self.run_thread(thread)
            thread_id = thread.id
        else:
            thread = self._get_thread(thread_id)
            self.continue_thread(prompt, thread_id)
            run = self.run_thread(thread)
        while run.status in {"queued", "in_progress"}:
            run = self.client.beta.threads.runs.retrieve(
                thread_id=thread_id,
                run_id=run.id,
            )
            await asyncio.sleep(0.5)
        return run

    async def image_prompt(self, prompt: str) -> Optional[str]:
        """
        Request an image to be generated using a separate image model.

        :param prompt: The image prompt.
        :return: The URL of the generated image.
        """
        self.last_prompt = prompt
        response = self.client.images.generate(
            model=environment.IMAGE_MODEL,
            prompt=prompt,
            size="1024x1024",
            quality="standard",
            n=1,
        )
        return response.data[0].url

    def get_last_message(self, thread_id: str) -> Optional[MessageData]:
        messages = self.client.beta.threads.messages.list(
            thread_id=thread_id,
            order="asc",
            after=self.last_message.id if self.last_message else NOT_GIVEN,
        ).data

        last_message_in_thread = messages[-1]

        if not last_message_in_thread:
            return None

        if self.last_message and last_message_in_thread.id == self.last_message.id:
            raise NoResponseError

        self.last_message = last_message_in_thread

        return MessageData(
            text_content=last_message_in_thread.content[0].text.value,
            thread_id=thread_id,
        )

    async def converse(
        self, user_input: str, thread_id: Optional[str] = None
    ) -> Optional[MessageData]:
        """
        Converse with the assistant by creating or continuing a thread.

        :param user_input: The user's input message.
        :param thread_id: Optional ID of the thread to continue.
        :return: The last message in the thread.
        """
        if not user_input:
            return None

        if thread_id is None:
            run = await self.prompt(user_input)
            thread_id = run.thread_id
        else:
            await self.prompt(user_input, thread_id)

        return self.get_last_message(thread_id)

    async def save_conversation_state(self) -> str:
        """
        Save the state of the conversation.
        :return: The thread ID of the conversation.
        """
        await threads_table.save_thread(
            self.last_message.thread_id, self.assistant_id, self.last_prompt
        )
        return self.last_message.thread_id


class Completion(MemoryMixin):
    """
    Encapsulates interactions with the OpenAI Chat Completion API.

    Inherits from:
        - AssistantProtocol: Protocol defining the interface for assistant classes.
        - MemoryMixin: Mixin class to handle memory-related functionality.

    Attributes:
        model (str): The model to be used for completions.
        max_memory (int): Maximum number of messages to retain in memory.
        api_key (str): API key for OpenAI.
        client (openai.OpenAI): Client for interacting with the OpenAI API.
    """

    REASONING_MODELS = ["o1", "o3-mini"]

    def __init__(
        self,
        model: str,
        max_memory: int = 50,
        api_key: str = environment.OPENAI_API_KEY,
        reasoning_effort: str = "medium",
    ):
        """
        Initialize the Completion instance.

        :param model: The model to be used for completions.
        :param max_memory: Maximum number of messages to retain in memory.
        :param api_key: API key for OpenAI.
        """
        if not api_key:
            raise ConfigError("Missing 'OPENAI_API_KEY' environment variable")

        MemoryMixin.__init__(self, max_memory)
        self.client = openai.OpenAI(api_key=api_key)
        self.model = model
        self.reasoning_effort = (
            reasoning_effort if self.model in self.REASONING_MODELS else None
        )

    async def start(self):
        """
        Do nothing
        """
        pass

    def complete(self, prompt: str) -> ChatCompletionMessage:
        """
        Generate a completion for the given prompt.

        :param prompt: The prompt to complete.
        :return: The completion message.
        """
        new_prompt = MessageDict(
            role="user",
            content=prompt,
        )
        self.remember(new_prompt)
        response = self.client.chat.completions.create(
            model=self.model,
            messages=self.memory,  # type: ignore
            reasoning_effort=self.reasoning_effort,
        )
        message = response.choices[0].message
        self.remember({"role": "assistant", "content": message.content})
        return response.choices[0].message

    async def converse(
        self, user_input: str, *args, **kwargs  # pylint: disable=unused-argument
    ) -> Optional[MessageData]:
        """
        Converse with the assistant using the chat completion API.

        :param user_input: The user's input message.
        :return: The completion message.
        """
        if not user_input:
            return None

        message = self.complete(user_input)
        return MessageData(text_content=message.content or "")



import os
from pathlib import Path

HOME_DIR = Path.home()

_override_config_dir = os.getenv("ASSISTANTS_CONFIG_DIR")

CONFIG_DIR = (
    Path(_override_config_dir)
    if _override_config_dir
    else HOME_DIR / ".config" / "assistants"
)

if not CONFIG_DIR.exists():
    CONFIG_DIR.mkdir(parents=True)

_override_data_dir = os.getenv("ASSISTANTS_DATA_DIR")

DATA_DIR = (
    Path(_override_data_dir)
    if _override_data_dir
    else HOME_DIR / ".local" / "share" / "assistants"
)

DB_PATH = DATA_DIR / "data.db"



from functools import wraps

from telegram import Update

from assistants.user_data.interfaces.telegram_chat_data import NotAuthorized
from assistants.user_data.sqlite_backend.telegram_chat_data import (
    TelegramSqliteUserData,
)

chat_data = TelegramSqliteUserData()


def restricted_access(f):
    @wraps(f)
    async def wrapper(update: Update, *args, **kwargs):
        try:
            await chat_data.check_chat_authorised(update.effective_chat.id)
        except NotAuthorized:
            await chat_data.check_user_authorised(update.effective_user.id)

        return await f(update, *args, **kwargs)

    return wrapper


def requires_superuser(f):
    @wraps(f)
    async def wrapper(update: Update, *args, **kwargs):
        await chat_data.check_superuser(update.effective_user.id)

        return await f(update, *args, **kwargs)

    return wrapper



import sys


class ANSIEscapeSequence:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"  # (e.g. Orange)
    FAIL = "\033[91m"  # (e.g. Red)
    ENDC = "\033[0m"  # Reset to default
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    # Clear the screen and move the cursor to the top left corner
    CLEAR_SCREEN = "\033[2J\033[H"


def clear_screen():
    """Clear the terminal screen, terminal must be ANSI escape sequence compatible"""
    sys.stdout.write(ANSIEscapeSequence.CLEAR_SCREEN)
    sys.stdout.flush()



"""
This module defines the types used by the generic assistants API.
These components are used to represent and manage message data and interactions with assistant classes.

Classes:
    - MessageData: Data class representing message data.
    - AssistantProtocol: Protocol defining the interface for assistant classes.
    - MessageDict: Typed dictionary for message data.
"""

from dataclasses import dataclass
from typing import Optional, Protocol, TypedDict


@dataclass
class MessageData:
    """
    Data class representing message data.

    Attributes:
        text_content (str): The text content of the message.
        thread_id (Optional[str]): The ID of the thread the message belongs to.
    """

    text_content: str
    thread_id: Optional[str] = None


class AssistantProtocol(Protocol):
    """
    Protocol defining the interface for assistant classes.

    Attributes:
        assistant_id (Optional[str]): The ID of the assistant.
    """

    assistant_id: Optional[str]

    async def start(self) -> None:
        """
        Load the assistant etc. (async init if needed).
        """
        ...

    async def converse(
        self, user_input: str, thread_id: Optional[str] = None
    ) -> Optional[MessageData]:
        """
        Converse with the assistant.

        :param user_input: The user's input message.
        :param thread_id: Optional ID of the thread to continue.
        :return: The last message in the thread or a string response.
        """
        ...

    def get_last_message(self, thread_id: str) -> Optional[MessageData]:
        """
        Get the last message in the thread.

        :param thread_id: the ID of the thread to continue.
        :return: last message in the thread if one exists.
        """
        ...

    def save_conversation_state(self) -> str:
        """
        Save the current state of the conversation.
        :return: The conversation ID/ Thread ID.
        """
        ...


class MessageDict(TypedDict):
    """
    Typed dictionary for message data.

    Attributes:
        role (str): The role of the message sender (e.g., 'user', 'assistant').
        content (Optional[str]): The content of the message.
    """

    role: str
    content: str | None



from datetime import datetime
from pathlib import Path
from typing import NamedTuple, Optional

import aiosqlite

from assistants.config.file_management import DB_PATH

TABLE_NAME = "threads"


class ThreadData(NamedTuple):
    thread_id: str
    last_run_dt: Optional[str]
    assistant_id: Optional[str]
    initial_prompt: Optional[str]


class NewThreadData(NamedTuple):
    thread_id: str
    assistant_id: Optional[str]


class ThreadsTable:
    def __init__(self, db_path: Path = DB_PATH):
        self.db_path = db_path

    async def get_by_thread_id(self, thread_id: str) -> Optional[ThreadData]:
        async with aiosqlite.connect(self.db_path) as conn:
            cur = conn.cursor()
            cur.execute(
                "SELECT thread_id, last_run_dt, assistant_id, initial_prompt FROM threads WHERE thread_id = ?",
                (thread_id,),
            )
            row = cur.fetchone()
            if row is None:
                return None
            return ThreadData(*row)

    async def get_by_assistant_id(
        self, assistant_id: str, limit: int = 0
    ) -> list[ThreadData]:
        async with aiosqlite.connect(self.db_path) as conn:
            cur = await conn.cursor()
            statement = (
                "SELECT thread_id, last_run_dt, assistant_id, initial_prompt FROM threads WHERE assistant_id = ?"
                "ORDER BY last_run_dt DESC"
            )
            if limit > 0:
                statement = statement + " LIMIT ?"
                binding = (assistant_id, limit)
            else:
                binding = (assistant_id,)
            await cur.execute(
                statement,
                binding,
            )
            rows = await cur.fetchall()
            results = []
            for row in rows:
                results.append(ThreadData(*row))
            return results

    async def get_all_threads(self) -> list[ThreadData]:
        async with aiosqlite.connect(self.db_path) as conn:
            cur = await conn.cursor()
            await cur.execute(
                "SELECT thread_id, last_run_dt, assistant_id, initial_prompt FROM threads WHERE TRUE "
                "ORDER BY last_run_dt DESC;"
            )
            rows = await cur.fetchall()
            results = []
            for row in rows:
                results.append(ThreadData(*row))

    async def save_thread(
        self, thread_id: str, assistant_id: str, initial_prompt: Optional[str] = None
    ) -> None:
        async with aiosqlite.connect(self.db_path) as conn:
            # Select the initial prompt for the thread_id if it exists
            async with await conn.execute(
                f"SELECT initial_prompt FROM {TABLE_NAME} WHERE thread_id = '{thread_id}';"
            ) as cursor:
                result = await cursor.fetchone()
                initial_prompt = result[0] if result else initial_prompt
            await conn.execute(
                f"REPLACE INTO {TABLE_NAME} VALUES (?, ?, ?, ?);",
                (thread_id, assistant_id, datetime.now().isoformat(), initial_prompt),
            )
            await conn.commit()


threads_table = ThreadsTable()


async def get_last_thread_for_assistant(assistant_id: str):
    result = await threads_table.get_by_assistant_id(assistant_id, limit=1)
    if result:
        return result[0]



import json
import re
import webbrowser
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import UTC, datetime
from typing import Optional

import aiofiles
import aiohttp
import pyperclip

from assistants.ai.memory import MemoryMixin
from assistants.ai.openai import Assistant
from assistants.ai.types import AssistantProtocol, MessageData
from assistants.cli import output
from assistants.lib.constants import IO_INSTRUCTIONS
from assistants.cli.selector import TerminalSelector, TerminalSelectorOption
from assistants.cli.terminal import clear_screen
from assistants.cli.utils import get_text_from_default_editor, highlight_code_blocks
from assistants.config import environment
from assistants.config.file_management import DATA_DIR
from assistants.user_data import threads_table
from assistants.user_data.sqlite_backend import conversations_table


@dataclass
class IoEnviron:
    """
    Environment variables for the input/output loop.
    """

    assistant: AssistantProtocol | MemoryMixin | Assistant
    last_message: Optional[MessageData] = None
    thread_id: Optional[str] = None
    user_input: Optional[str] = None


class Command(ABC):
    """
    Command protocol for the input/output loop.
    """

    @abstractmethod
    async def __call__(self, environ: IoEnviron, *args) -> None:
        """
        Call the command.

        :param environ: The environment variables for the input/output loop.
        """
        pass


class Editor(Command):
    """
    Command to open the default text editor.
    """

    async def __call__(self, environ: IoEnviron, *args) -> None:
        """
        Call the command to open the default text editor.

        :param environ: The environment variables for the input/output loop.
        """
        environ.user_input = get_text_from_default_editor().strip()


editor: Command = Editor()


class CopyResponse(Command):
    """
    Command to copy the response to the clipboard.
    """

    @staticmethod
    def copy_to_clipboard(text: str) -> None:
        """
        Copy the given text to the clipboard.

        :param text: The text to copy to the clipboard.
        """
        try:
            pyperclip.copy(text)
        except pyperclip.PyperclipException:
            output.fail(
                "Error copying to clipboard; this feature doesn't seem to be "
                "available in the current terminal environment."
            )
            return

    @staticmethod
    def get_previous_response(environ: IoEnviron) -> str:
        """
        Get the previous response from the assistant.

        :param environ: The environment variables for the input/output loop.
        :return: The previous response from the assistant.
        """
        previous_response = ""

        if environ.last_message:
            previous_response = environ.last_message.text_content

        return previous_response

    async def __call__(self, environ: IoEnviron, *args) -> None:
        """
        Call the command to copy the response to the clipboard.

        :param environ: The environment variables for the input/output loop.
        """
        previous_response = self.get_previous_response(environ)

        if not previous_response:
            output.warn("No previous message to copy.")
            return

        # Check if previous response is a URL
        if re.match(
            r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+",
            previous_response,
        ):
            message = "image URL"
        else:
            message = "response"

        self.copy_to_clipboard(previous_response)
        output.inform(f"Copied {message} to clipboard")


copy_response: Command = CopyResponse()


class CopyCodeBlocks(CopyResponse):
    """
    Command to copy the code blocks from the response to the clipboard.
    """

    async def __call__(self, environ: IoEnviron, *args) -> None:
        """
        Call the command to copy the code blocks from the response to the clipboard.

        :param environ: The environment variables for the input/output loop.
        """
        previous_response = self.get_previous_response(environ)

        if not previous_response:
            output.warn("No previous message to copy from.")
            return

        split_code = re.split(r"(```.*?```)", previous_response, flags=re.DOTALL)

        pattern = r"```(?:[a-zA-Z]+(\n))?(\n)?([\s\S]*?)\n?```"
        replacement = r"\1\2\3\2\1"

        code_blocks = [
            re.sub(pattern, replacement, block)
            for block in split_code
            if block.startswith("```")
        ]

        if args:
            try:
                code_blocks = [code_blocks[int(str(args[0]))]]
            except (ValueError, IndexError):
                output.fail(
                    "Pass the index of the code block to copy, or no arguments to copy all code blocks."
                )
                return

        if not code_blocks:
            output.warn("No codeblocks in previous message!")
            return

        if code_blocks[0].startswith("\n"):
            code_blocks[0] = code_blocks[0][1:]  # Remove the leading newline
        if code_blocks[-1].endswith("\n"):
            code_blocks[-1] = code_blocks[-1][:-1]  # Remove the trailing newline

        if len(code_blocks) > 1:
            for i, block in enumerate(code_blocks[:-1]):
                if not block.endswith("\n"):
                    code_blocks[i] = block + "\n"

        all_code = "".join(code_blocks)

        self.copy_to_clipboard(all_code)

        output.inform(
            f"Copied code block{'s' if not args and len(code_blocks) > 1 else ''} to clipboard"
        )


copy_code_blocks: Command = CopyCodeBlocks()


class PrintUsage(Command):
    """
    Command to print the usage instructions.
    """

    async def __call__(self, environ: IoEnviron, *args) -> None:
        """
        Call the command to print the usage instructions.

        :param environ: The environment variables for the input/output loop.
        """
        output.inform(IO_INSTRUCTIONS)


print_usage: Command = PrintUsage()


class NewThread(Command):
    """
    Command to start a new thread.
    """

    async def __call__(self, environ: IoEnviron, *args) -> None:
        """
        Call the command to start a new thread.

        :param environ: The environment variables for the input/output loop.
        """
        environ.thread_id = None
        environ.last_message = None
        await environ.assistant.start()
        clear_screen()


new_thread: Command = NewThread()


class SelectThread(Command):
    """
    Command to select a thread.
    """

    async def __call__(self, environ: IoEnviron, *args) -> None:
        """
        Call the command to select a thread.

        :param environ: The environment variables for the input/output loop.
        """
        if isinstance(environ.assistant, MemoryMixin):
            threads = await conversations_table.get_all_conversations()

            thread_options = [
                TerminalSelectorOption(
                    label=f"{thread.last_updated} | {json.loads(thread.conversation)[0]['content']}",
                    value=thread.id,
                )
                for thread in threads
            ]
        else:
            threads = await threads_table.get_by_assistant_id(
                environ.assistant.assistant_id
            )
            thread_options = [
                TerminalSelectorOption(
                    label=f"{thread.last_run_dt} | {thread.initial_prompt}",
                    value=thread.thread_id,
                )
                for thread in threads
            ]

        if not threads:
            output.warn("No threads found.")
            return

        selector = TerminalSelector(
            thread_options, title="Select a thread to continue..."
        )
        thread_id = selector.run()
        if not thread_id:
            return  # No change

        if thread_id == environ.thread_id:
            return  # No change

        environ.thread_id = thread_id

        if isinstance(environ.assistant, MemoryMixin):
            await environ.assistant.load_conversation(thread_id)
        else:
            await environ.assistant.start()

        output.inform(f"Selected thread '{thread_id}'")

        last_message = environ.assistant.get_last_message(thread_id)
        environ.last_message = last_message

        if last_message:
            output.default(highlight_code_blocks(last_message.text_content))
            output.new_line(2)
        else:
            output.warn("No last message found in thread")


select_thread: Command = SelectThread()


class GenerateImage(Command):
    @staticmethod
    async def save_image(image_url: str, prompt: str) -> None:
        """
        Save the image URL to the database.

        :param image_url: The URL of the image to save.
        :param prompt: The prompt that was used to create the image.
        """
        # Get the image content
        async with aiohttp.ClientSession() as session:
            async with session.get(image_url) as response:
                image_content = await response.read()

        # Save the image to file
        image_path = DATA_DIR / "images"
        if not image_path.exists():
            image_path.mkdir(parents=True)

        filename = f"{'_'.join(prompt.split())}_{datetime.now(UTC).timestamp()}.png"
        image_path /= filename

        async with aiofiles.open(image_path, "wb") as file:
            await file.write(image_content)

        output.inform(f"Image saved to {image_path}")

    async def __call__(self, environ: IoEnviron, *args) -> None:
        assistant = environ.assistant
        if not isinstance(assistant, Assistant):
            raise NotImplementedError

        prompt = " ".join(args)

        image_url = await assistant.image_prompt(prompt)

        if image_url:
            output.default(f"Here's your image:\n{image_url}")
            output.new_line(2)

            environ.last_message = MessageData(
                text_content=image_url, thread_id=environ.thread_id
            )

            if environment.OPEN_IMAGES_IN_BROWSER:
                webbrowser.open(image_url)
                output.inform("Opening image in browser...")

            if input("Would you like to save this image? (y/N): ").lower() == "y":
                await self.save_image(image_url, prompt)

        else:
            output.warn("No image returned...")


generate_image: Command = GenerateImage()

COMMAND_MAP = {
    "/e": editor,
    "/edit": editor,
    "/editor": editor,
    "/c": copy_response,
    "/copy": copy_response,
    "/cc": copy_code_blocks,
    "/copy-code": copy_code_blocks,
    "/h": print_usage,
    "/help": print_usage,
    "/n": new_thread,
    "/new": new_thread,
    "/new-thread": new_thread,
    "/t": select_thread,
    "/threads": select_thread,
    "/i": generate_image,
    "/image": generate_image,
}

EXIT_COMMANDS = {
    "q",
    "quit",
    "exit",
    "/q",
    "/quit",
    "/exit",
}



from typing import NamedTuple, Optional

import aiosqlite

from assistants.config.file_management import DB_PATH

TABLE_NAME = "assistants"


class AssistantData(NamedTuple):
    assistant_id: Optional[str]
    config_hash: Optional[str]


async def get_assistant_data(assistant_name: str, config_hash: str):
    async with aiosqlite.connect(DB_PATH) as db:
        async with await db.execute(
            f"SELECT assistant_id, config_hash FROM {TABLE_NAME} WHERE assistant_name = '{assistant_name}';"
        ) as cursor:
            result = await cursor.fetchone()
            if result:
                return AssistantData(*result)

        await db.execute(
            f"REPLACE INTO {TABLE_NAME} VALUES ('{assistant_name}', NULL, '{config_hash}');"
        )
        await db.commit()
        return AssistantData(None, config_hash)


async def save_assistant_id(assistant_name: str, assistant_id: str, config_hash: str):
    async with aiosqlite.connect(DB_PATH) as db:
        await db.execute(
            f"REPLACE INTO {TABLE_NAME} VALUES ('{assistant_name}', '{assistant_id}', '{config_hash}');"
        )
        await db.commit()



IO_INSTRUCTIONS = """\
Commands:

/h,  /help      Show this help message
/e,  /editor    Open the default editor to compose a prompt
/i,  /image <prompt>     
                Generate an image from the prompt supplied as args
                e.g. `/i a dog riding a pony, in the style of Botero`
/c,  /copy      Copy the previous response to the clipboard
/cc, /copy-code [i]
                Copy the code blocks from the previous response to the clipboard
                (an optional index can be supplied to copy a single code block)
/n,  /new       Start a new thread and clear the terminal screen
/t,  /threads:  List all the threads, and select one to continue
/clear, C-l     Clear the terminal screen without starting a new thread

Press Ctrl+C or Ctrl+D to exit the program

Anything else you type will be sent to the assistant for processing.\
"""

CLAUDE_CLI_MAX_TOKENS = 2048



import os
import re
import subprocess
import tempfile
from argparse import Namespace
from typing import Optional, cast

from pygments import highlight
from pygments.formatters import TerminalFormatter
from pygments.lexers import get_lexer_by_name

from assistants.ai.anthropic import Claude
from assistants.ai.dummy_assistant import DummyAssistant
from assistants.ai.openai import Assistant, Completion
from assistants.ai.types import AssistantProtocol
from assistants.cli import output
from assistants.config import environment
from assistants.lib.exceptions import ConfigError


def highlight_code_blocks(markdown_text):
    code_block_pattern = re.compile(r"```(\w+)?\n(.*?)```", re.DOTALL)

    def replacer(match):
        lang = match.group(1)
        code = match.group(2)
        if lang:
            if lang == "plaintext":
                lang = "text"
            lexer = get_lexer_by_name(lang, stripall=True)
        else:
            lexer = get_lexer_by_name("text", stripall=True)
        return f"```{lang if lang else ''}\n{highlight(code, lexer, TerminalFormatter())}```"

    return code_block_pattern.sub(replacer, markdown_text)


def get_text_from_default_editor(initial_text=None):
    # Create a temporary file
    with tempfile.NamedTemporaryFile(suffix=".md", delete=False) as temp_file:
        temp_file_path = temp_file.name

    if initial_text:
        with open(temp_file_path, "w") as text_file:
            text_file.write(initial_text)

    # Open the editor for the user to input text
    editor = os.environ.get("EDITOR", "nano")
    subprocess.run([editor, temp_file_path])

    # Read the contents of the file after the editor is closed
    with open(temp_file_path, "r") as file:
        text = file.read()

    # Remove the temporary file
    os.remove(temp_file_path)

    return text


MODEL_LOOKUP = {
    "code": {
        "o1": Completion,
        "o3": Completion,
        "claude-": Claude,
    },
    "default": {
        "claude-": Claude,
        "dummy-model": DummyAssistant,
        "gpt-4o": Assistant,
        "o1": Assistant,
        "o3": Assistant,
    },
}


async def create_assistant_and_thread(
    args: Namespace,
) -> tuple[AssistantProtocol, Optional[str]]:
    thread_id = None

    def get_model_class(model_type: str, model_name: str):
        for key, assistant_type in MODEL_LOOKUP[model_type].items():
            if model_name.startswith(key):
                return assistant_type
        raise ConfigError(f"Invalid {model_type} model: {model_name}")

    if args.code:
        model_class = get_model_class("code", environment.CODE_MODEL)
        assistant = model_class(model=environment.CODE_MODEL)
        if isinstance(assistant, Claude):
            assistant.thinking = True

    else:
        if args.instructions:
            try:
                with open(args.instructions, "r") as instructions_file:
                    instructions_text = instructions_file.read()
            except FileNotFoundError:
                raise ConfigError(f"Instructions file not found: '{args.instructions}'")
        else:
            instructions_text = environment.ASSISTANT_INSTRUCTIONS

        model_class = get_model_class("default", environment.DEFAULT_MODEL)

        if model_class == Assistant:
            assistant = model_class(
                name=environment.ASSISTANT_NAME,
                model=environment.DEFAULT_MODEL,
                instructions=instructions_text,
                tools=[{"type": "code_interpreter"}],
            )
        elif model_class == Claude:
            assistant = model_class(
                model=environment.DEFAULT_MODEL, instructions=instructions_text
            )
        else:
            assistant = model_class(model=environment.DEFAULT_MODEL)

        if instructions_text and isinstance(assistant, Claude):
            output.warn(
                "Custom instructions are not fully supported with this assistant."
            )

    await assistant.start()
    if args.continue_thread:
        thread_id = await assistant.async_get_conversation_id()

    assistant = cast(AssistantProtocol, assistant)

    return assistant, thread_id



import os

import aiohttp
from telegram import Update
from telegram.constants import ParseMode
from telegram.ext import ContextTypes

from assistants.telegram_ui.auth import (
    restricted_access,
    requires_superuser,
    chat_data,
)
from assistants.telegram_ui.lib import requires_reply_to_message, assistant
from assistants.user_data import threads_table
from assistants.user_data.interfaces.telegram_chat_data import ChatHistory


@requires_superuser
@requires_reply_to_message
async def promote_user(update: Update, context: ContextTypes.DEFAULT_TYPE):
    await chat_data.promote_superuser(update.message.reply_to_message.from_user.id)
    await context.bot.send_message(
        chat_id=update.effective_chat.id, text="User promoted"
    )


@requires_superuser
@requires_reply_to_message
async def demote_user(update: Update, context: ContextTypes.DEFAULT_TYPE):
    await chat_data.demote_superuser(update.message.reply_to_message.from_user.id)
    await context.bot.send_message(
        chat_id=update.effective_chat.id, text="User demoted"
    )


@requires_superuser
async def authorise_chat(update: Update, context: ContextTypes.DEFAULT_TYPE):
    await chat_data.authorise_chat(update.effective_chat.id)
    await context.bot.send_message(
        chat_id=update.effective_chat.id, text="Chat authorised"
    )


@requires_superuser
@requires_reply_to_message
async def authorise_user(update: Update, context: ContextTypes.DEFAULT_TYPE):
    await chat_data.authorise_chat(update.message.reply_to_message.from_user.id)
    await context.bot.send_message(
        chat_id=update.effective_chat.id, text="User authorised"
    )


@requires_superuser
async def deauthorise_chat(update: Update, context: ContextTypes.DEFAULT_TYPE):
    await chat_data.deauthorise_chat(update.effective_chat.id)
    await context.bot.send_message(
        chat_id=update.effective_chat.id, text="Chat de-authorised"
    )


@requires_superuser
@requires_reply_to_message
async def deauthorise_user(update: Update, context: ContextTypes.DEFAULT_TYPE):
    await chat_data.deauthorise_user(update.message.reply_to_message.from_user.id)
    await context.bot.send_message(
        chat_id=update.effective_chat.id, text="User de-authorised"
    )


@restricted_access
async def new_thread(update: Update, context: ContextTypes.DEFAULT_TYPE):
    await chat_data.clear_last_thread_id(update.effective_chat.id)
    assistant.last_message = None
    await context.bot.send_message(
        update.effective_chat.id, "Conversation history cleared."
    )


@restricted_access
async def toggle_auto_reply(update: Update, context: ContextTypes.DEFAULT_TYPE):
    chat_history = await chat_data.get_chat_history(update.effective_chat.id)
    result = "OFF" if chat_history.auto_reply else "ON"
    await chat_data.set_auto_reply(
        update.effective_chat.id, not chat_history.auto_reply
    )
    await context.bot.send_message(
        chat_id=update.effective_chat.id, text=f"Auto reply is {result}"
    )


@restricted_access
async def message_handler(update: Update, context: ContextTypes.DEFAULT_TYPE):
    existing_chat = await chat_data.get_chat_history(update.effective_chat.id)
    message_text = update.message.text
    if not existing_chat.auto_reply:
        bot_id = context.bot.id
        bot_username = f"@{context.bot.username}"
        if bot_username not in message_text and (
            not update.message.reply_to_message
            or update.message.reply_to_message.from_user.id != bot_id
        ):
            return
        message_text = message_text.replace(
            bot_username, os.getenv("ASSISTANT_NAME", "[ASSISTANT NAME]")
        )

    response_message = await assistant.converse(message_text, existing_chat.thread_id)

    if not existing_chat.thread_id:
        await chat_data.save_chat_history(
            ChatHistory(
                chat_id=update.effective_chat.id,
                thread_id=response_message.thread_id,
                auto_reply=existing_chat.auto_reply,
            )
        )
        await threads_table.save_thread(
            response_message.thread_id, assistant.assistant_id, message_text
        )

    response = response_message.text_content

    response_parts = response.split("```")

    if len(response_parts) % 2 == 0:
        # Should be an odd number of parts if codeblocks
        await context.bot.send_message(
            chat_id=update.effective_chat.id,
            text=response,
        )
        return

    for i, part in enumerate(response_parts):
        if i % 2:
            await context.bot.send_message(
                chat_id=update.effective_chat.id,
                text=f"```{part}```",
                parse_mode=ParseMode.MARKDOWN_V2,
            )
        else:
            await context.bot.send_message(
                chat_id=update.effective_chat.id,
                text=part,
            )


@restricted_access
async def generate_image(update: Update, context: ContextTypes.DEFAULT_TYPE):
    prompt = update.message.text.replace("/image ", "")
    image_url = await assistant.image_prompt(prompt)

    async with aiohttp.ClientSession() as session:
        async with session.get(image_url) as response:
            image_content = await response.read()

    await update.message.reply_photo(image_content)



"""
This module contains the main input/output loop for interacting with the assistant.
"""

import asyncio
import sys
from dataclasses import dataclass
from enum import Enum
from typing import Optional

from prompt_toolkit import prompt
from prompt_toolkit.history import FileHistory
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.styles import Style

from assistants.ai.memory import MemoryMixin
from assistants.ai.types import AssistantProtocol
from assistants.cli import output
from assistants.cli.commands import COMMAND_MAP, EXIT_COMMANDS, IoEnviron
from assistants.cli.terminal import clear_screen
from assistants.cli.utils import highlight_code_blocks
from assistants.config.file_management import CONFIG_DIR
from assistants.log import logger


# Constants and Configuration
class PromptStyle(Enum):
    USER_INPUT = "ansigreen"
    PROMPT_SYMBOL = "ansibrightgreen"


INPUT_CLASSNAME = "input"


@dataclass
class PromptConfig:
    style: Style = Style.from_dict(
        {
            "": PromptStyle.USER_INPUT.value,
            INPUT_CLASSNAME: PromptStyle.PROMPT_SYMBOL.value,
        }
    )
    prompt_symbol: str = ">>>"
    history_file: str = f"{CONFIG_DIR}/history"


# Setup
bindings = KeyBindings()
config = PromptConfig()
history = FileHistory(config.history_file)
PROMPT = [(f"class:{INPUT_CLASSNAME}", f"{config.prompt_symbol} ")]


# Bind CTRL+L to clear the screen
@bindings.add("c-l")
def _(_event):
    clear_screen()


def get_user_input() -> str:
    """Get user input from interactive/styled prompt (prompt_toolkit)."""
    if not sys.stdin.isatty():
        sys.stdin = open("/dev/tty")
    return prompt(PROMPT, style=config.style, history=history, in_thread=True)


async def io_loop_async(
    assistant: AssistantProtocol | MemoryMixin,
    initial_input: str = "",
    thread_id: Optional[str] = None,
):
    """
    Main input/output loop for interacting with the assistant.

    :param assistant: The assistant instance implementing AssistantProtocol.
    :param initial_input: Initial user input to start the conversation.
    :param thread_id: The ID of the conversation thread.
    """
    environ = IoEnviron(
        assistant=assistant,
        thread_id=thread_id,
    )
    while (
        initial_input or (user_input := get_user_input()).lower() not in EXIT_COMMANDS
    ):
        output.reset()
        environ.user_input = None
        if initial_input:
            output.user_input(initial_input)
            user_input = initial_input
            initial_input = ""  # Otherwise, the initial input will be repeated in the next iteration

        user_input = user_input.strip()

        if not user_input:
            continue

        # Handle commands
        c, *args = user_input.split(" ")
        command = COMMAND_MAP.get(c.lower())
        if command:
            logger.debug(
                f"Command input: {user_input}; Command: {command.__class__.__name__}"
            )
            await command(environ, *args)
            if environ.user_input:
                initial_input = environ.user_input
            continue

        if user_input.startswith("/"):
            output.warn("Invalid command!")
            continue

        environ.user_input = user_input
        await converse(environ)


async def converse(
    environ: IoEnviron,
):
    """
    Handle the conversation with the assistant.

    :param environ: The environment variables manipulated on each
    iteration of the input/output loop.
    """
    assistant = environ.assistant
    last_message = environ.last_message
    thread_id = environ.thread_id  # Could be None; a new thread will be created if so.

    message = await assistant.converse(
        environ.user_input, last_message.thread_id if last_message else thread_id
    )

    if (
        message is None
        or not message.text_content
        or last_message
        and last_message.text_content == message.text_content
    ):
        output.warn("No response from the AI model.")
        return

    text = highlight_code_blocks(message.text_content)

    output.default(text)
    output.new_line(2)

    # Set and save the new conversation state for future iterations:
    environ.last_message = message
    environ.thread_id = await assistant.save_conversation_state()


def io_loop(
    assistant: AssistantProtocol | MemoryMixin,
    initial_input: str = "",
    thread_id: Optional[str] = None,
):
    asyncio.run(io_loop_async(assistant, initial_input, thread_id))



"""
This module provides the `MemoryMixin` class, which handles memory-related functionality for managing conversations.

Classes:
    - MemoryMixin: Mixin class to handle memory-related functionality, including remembering messages, truncating memory, and loading/saving conversations from/to a database.
"""

import json
import uuid
from datetime import datetime
from typing import Optional

from assistants.ai.types import MessageData, MessageDict
from assistants.user_data.sqlite_backend import conversations_table
from assistants.user_data.sqlite_backend.conversations import Conversation


class MemoryMixin:
    """
    Mixin class to handle memory-related functionality.
    """

    def __init__(self, max_memory: int = 50):
        """
        Initialize the MemoryMixin instance.

        :param max_memory: Maximum number of messages to retain in memory.
        """
        self.memory: list[MessageDict] = []
        self.max_memory = max_memory
        self.conversation_id = None

    def truncate_memory(self):
        """
        Truncate the memory to the maximum allowed messages.
        """
        self.memory = self.memory[-self.max_memory :]

    def remember(self, message: MessageDict):
        """
        Remember a new message.

        :param message: The message to remember.
        """
        self.truncate_memory()
        self.memory.append(message)

    async def load_conversation(self, conversation_id: Optional[str] = None):
        """
        Load the last conversation from the database.

        :param conversation_id: Optional ID of the conversation to load.
        """
        if conversation_id:
            conversation = await conversations_table.get_conversation(conversation_id)
        else:
            conversation = await conversations_table.get_last_conversation()

        self.memory = json.loads(conversation.conversation) if conversation else []
        self.conversation_id = conversation.id if conversation else uuid.uuid4().hex

    async def async_get_conversation_id(self):
        if not self.conversation_id:
            await self.load_conversation()
        return self.conversation_id

    async def save_conversation_state(self) -> str:
        """
        Save the current conversation to the database.
        :return: The conversation ID.
        """
        if not self.memory:
            return

        if self.conversation_id is None:
            self.conversation_id = uuid.uuid4().hex

        await conversations_table.save_conversation(
            Conversation(
                id=self.conversation_id,
                conversation=json.dumps(self.memory),
                last_updated=datetime.now(),
            )
        )
        return self.conversation_id

    def get_last_message(self, thread_id: str) -> Optional[MessageData]:
        """
        Get the last message from the conversation or None if no message exists.
        Conversation must have already been loaded.

        :param thread_id: Not used; required by protocol
        :return: MessageData with the last message and current conversation_id.
        """
        if not self.memory:
            return None
        return MessageData(
            text_content=self.memory[-1]["content"], thread_id=self.conversation_id
        )



import asyncio
import sys

from assistants.cli import cli
from assistants.config import environment
from assistants.user_data.sqlite_backend import init_db

CLAUDE_MODEL = "claude-3-5-sonnet-latest"


def main():
    if not environment.ANTHROPIC_API_KEY:
        print("ANTHROPIC_API_KEY not set in environment variables.", file=sys.stderr)
        sys.exit(1)
    environment.DEFAULT_MODEL = CLAUDE_MODEL
    environment.CODE_MODEL = CLAUDE_MODEL
    environment.ASSISTANT_INSTRUCTIONS = ""
    asyncio.run(init_db())
    cli()


if __name__ == "__main__":
    main()



"""
This module defines the Claude class, which encapsulates interactions with the
Anthropic API
It includes memory management functionality through the MemoryMixin class.

Classes:
    - Claude: Encapsulates interactions with the Anthropic API.
"""

from typing import Optional

from assistants.ai.memory import MemoryMixin
from assistants.ai.types import MessageData
from assistants.config import environment
from assistants.lib.exceptions import ConfigError

from anthropic import AsyncAnthropic

INSTRUCTIONS_UNDERSTOOD = "Instructions understood."


class Claude(MemoryMixin):
    """
    Claude class encapsulates interactions with the Anthropic API.

    Inherits from:
        - AssistantProtocol: Protocol defining the interface for assistant classes.
        - MemoryMixin: Mixin class to handle memory-related functionality.

    Attributes:
        model (str): The model to be used by the assistant.
        max_tokens (int): Maximum number of tokens for the response.
        max_memory (int): Maximum number of messages to retain in memory.
        client (AsyncAnthropic): Client for interacting with the Anthropic API.
    """

    def __init__(
        self,
        model: str,
        instructions: Optional[str] = None,
        max_tokens: int = environment.CLAUDE_MAX_TOKENS,
        max_memory: int = 50,
        api_key: Optional[str] = environment.ANTHROPIC_API_KEY,
        thinking: bool = False,
    ) -> None:
        """
        Initialize the Claude instance.

        :param model: The model to be used by the assistant.
        :param max_tokens: Maximum number of tokens for the response.
        :param max_memory: Maximum number of messages to retain in memory.
        :param api_key: API key for Anthropic. Defaults to ANTHROPIC_API_KEY.
        :raises ConfigError: If the API key is missing.
        """
        if not api_key:
            raise ConfigError("Missing 'ANTHROPIC_API_KEY' environment variable")

        MemoryMixin.__init__(self, max_memory)
        self.client = AsyncAnthropic(api_key=api_key)
        self.model = model
        self.max_tokens = max_tokens
        self.instructions = instructions
        self.thinking = thinking

    async def start(self) -> None:
        """
        Do nothing
        """
        await self.load_conversation()

    async def load_conversation(self, conversation_id: Optional[str] = None):
        """
        Load the conversation from the database.
        Also adds the instructions to the memory if provided and not
        already present, or not the most recent instructions.

        :param conversation_id: Optional ID of the conversation to load.
        """
        await super().load_conversation(conversation_id)
        if self.instructions:
            # Check if the instructions are already in memory
            for idx, message in enumerate(self.memory):
                if (
                    message.get("role") == "user"
                    and message.get("content") == self.instructions
                ):
                    understood_count = sum(
                        1
                        for msg in self.memory[idx + 1 :]
                        if msg.get("role") == "assistant"
                        and msg.get("content") == INSTRUCTIONS_UNDERSTOOD
                    )
                    if understood_count < 2:
                        # Most recent instructions are equivalent to the current ones
                        return

            self.memory = [
                *self.memory,
                {"role": "user", "content": self.instructions},
                {"role": "assistant", "content": INSTRUCTIONS_UNDERSTOOD},
            ]

    async def converse(
        self, user_input: str, thread_id: Optional[str] = None
    ) -> Optional[MessageData]:
        """
        Converse with the assistant by creating or continuing a thread.

        :param user_input: The user's input message.
        :param thread_id: Optional ID of the thread to continue.
        :return: The last message in the thread.
        """
        if not user_input:
            return None

        self.remember({"role": "user", "content": user_input})

        kwargs = {
            "max_tokens": self.max_tokens,
            "model": self.model,
            "messages": self.memory,
        }

        if self.thinking:
            kwargs["thinking"] = {
                "type": "enabled",
                "budget_tokens": (self.max_tokens // 4) * 3,
            }

        response = await self.client.messages.create(**kwargs)
        self.remember({"role": "assistant", "content": response.content[0].text})
        return MessageData(text_content=response.content[0].text)



"""
This module defines the `Conversation` data class and the `ConversationsTable` class for managing conversation records in an SQLite database.

Classes:
    - Conversation: Data class representing a conversation record.
    - ConversationsTable: Class for interacting with the conversations table in the SQLite database.
"""

from dataclasses import dataclass
from datetime import datetime
from typing import Optional

import aiosqlite

from assistants.config.file_management import DB_PATH


@dataclass
class Conversation:
    """
    Data class representing a conversation record.

    Attributes:
        id (str): The unique identifier of the conversation.
        conversation (str): The conversation data in JSON format.
        last_updated (datetime): The timestamp of the last update to the conversation.
    """

    id: str
    conversation: str
    last_updated: datetime


class ConversationsTable:
    """
    Class for interacting with the conversations table in the SQLite database.

    Attributes:
        db_path (str): The path to the SQLite database file.
    """

    def __init__(self, db_path: str):
        """
        Initialize the ConversationsTable instance.

        :param db_path: The path to the SQLite database file.
        """
        self.db_path = db_path

    async def create_table(self):
        """
        Create the conversations table if it does not exist.
        """
        async with aiosqlite.connect(self.db_path) as db:
            await db.execute(
                """
                CREATE TABLE IF NOT EXISTS conversations (
                    id TEXT PRIMARY KEY,
                    conversation TEXT,
                    last_updated TEXT
                )
            """
            )
            await db.commit()

    async def get_conversation(self, conversation_id: str) -> Optional[Conversation]:
        """
        Retrieve a conversation by its ID.

        :param conversation_id: The unique identifier of the conversation.
        :return: The Conversation object if found, otherwise None.
        """
        async with aiosqlite.connect(self.db_path) as db:
            cursor = await db.execute(
                """
                SELECT id, conversation, last_updated FROM conversations WHERE id = ?
            """,
                (conversation_id,),
            )
            row = await cursor.fetchone()
            if row:
                return Conversation(
                    id=row[0],
                    conversation=row[1],
                    last_updated=datetime.fromisoformat(row[2]),
                )

    async def save_conversation(self, conversation: Conversation):
        """
        Save a conversation to the database.

        :param conversation: The Conversation object to save.
        """
        async with aiosqlite.connect(self.db_path) as db:
            await db.execute(
                """
                REPLACE INTO conversations (id, conversation, last_updated) VALUES (?, ?, ?)
            """,
                (
                    conversation.id,
                    conversation.conversation,
                    conversation.last_updated.isoformat(),
                ),
            )
            await db.commit()

    async def delete_conversation(self, conversation_id: str):
        """
        Delete a conversation from the database by its ID.

        :param conversation_id: The unique identifier of the conversation to delete.
        """
        async with aiosqlite.connect(self.db_path) as db:
            await db.execute(
                """
                DELETE FROM conversations WHERE id = ?
            """,
                (conversation_id,),
            )
            await db.commit()

    async def get_all_conversations(self) -> list[Conversation]:
        """
        Retrieve all conversations from the database.

        :return: A list of Conversation objects.
        """
        async with aiosqlite.connect(self.db_path) as db:
            cursor = await db.execute(
                """
                SELECT id, conversation, last_updated FROM conversations ORDER BY last_updated DESC
            """
            )
            rows = await cursor.fetchall()
            result = []
            for row in rows:
                result.append(
                    Conversation(
                        id=row[0],
                        conversation=row[1],
                        last_updated=datetime.fromisoformat(row[2]),
                    )
                )
            return result

    async def get_last_conversation(self) -> Optional[Conversation]:
        """
        Retrieve the most recently updated conversation from the database.

        :return: The most recently updated Conversation object if found, otherwise None.
        """
        async with aiosqlite.connect(self.db_path) as db:
            cursor = await db.execute(
                """
                SELECT id, conversation, last_updated FROM conversations ORDER BY last_updated DESC LIMIT 1
            """
            )
            row = await cursor.fetchone()
            if row:
                return Conversation(
                    id=row[0],
                    conversation=row[1],
                    last_updated=datetime.fromisoformat(row[2]),
                )


conversations_table = ConversationsTable(DB_PATH)



from telegram.ext import (
    Application,
    ApplicationBuilder,
    CommandHandler,
    MessageHandler,
    filters,
)

from assistants.log import logger
from assistants.telegram_ui.commands import (
    promote_user,
    demote_user,
    authorise_chat,
    authorise_user,
    deauthorise_chat,
    deauthorise_user,
    new_thread,
    toggle_auto_reply,
    message_handler,
    generate_image,
)
from assistants.telegram_ui.lib import assistant


def build_bot(token: str) -> Application:
    application = ApplicationBuilder().token(token).build()
    application.add_handler(CommandHandler("add_chat", authorise_chat))
    application.add_handler(CommandHandler("remove_chat", deauthorise_chat))
    application.add_handler(CommandHandler("add_user", authorise_user))
    application.add_handler(CommandHandler("remove_user", deauthorise_user))
    application.add_handler(CommandHandler("promote", promote_user))
    application.add_handler(CommandHandler("demote", demote_user))
    application.add_handler(CommandHandler("new_thread", new_thread))
    application.add_handler(CommandHandler("auto_reply", toggle_auto_reply))
    application.add_handler(CommandHandler("image", generate_image))
    application.add_handler(MessageHandler(filters.TEXT, message_handler))
    logger.info("Setup complete!")
    return application


def run_polling(application: Application):
    logger.info("Telegram bot is running...")
    application.run_polling()


async def async_setup():
    await assistant.start()


def setup_and_run(token: str):
    application = build_bot(token)
    run_polling(application)



from assistants.cli.terminal import ANSIEscapeSequence


def reset():
    print(ANSIEscapeSequence.ENDC, end="")


def new_line(n=1):
    print("\n" * n, end="")


def green(text: str):
    print(f"{ANSIEscapeSequence.OKGREEN}{text if text else ''}", end="")


def warning(text: str):
    print(f"{ANSIEscapeSequence.WARNING}{text if text else ''}", end="")


def info(text: str):
    print(f"{ANSIEscapeSequence.OKBLUE}{text if text else ''}", end="")


def error(text: str):
    print(f"{ANSIEscapeSequence.FAIL}{text if text else ''}", end="")


def default(text: str):
    print(f"{ANSIEscapeSequence.ENDC}{text if text else ''}", end="")


def warn(text: str):
    warning(text)
    new_line(2)
    reset()


def fail(text: str):
    error(text)
    new_line(2)
    reset()


def inform(text: str):
    info(text)
    new_line(2)
    reset()


def user_input(text: str, prompt: str = ">>>"):
    green(f"{prompt} {text}")
    new_line()
    reset()



import asyncio
import os
import sys
from pathlib import Path

from assistants.cli import cli
from assistants.user_data.sqlite_backend import init_db


def install():
    # Get the path to the current environment's bin directory
    bin_dir = Path(sys.prefix) / "bin"
    path_update = f"export PATH=$PATH:{bin_dir}\n"

    # Check that the bin directory is not in the PATH environment variable already
    # and has not already been added to the path by this script
    path = os.environ.get("PATH", "")
    if str(bin_dir) in path:
        print(f"{bin_dir} is already in PATH")
        return

    with open(Path.home() / ".profile", "r") as f:
        if path_update in f.read():
            print(f"{bin_dir} is already in PATH")
            return

    # Add the bin directory to the PATH environment variable in .profile
    print(f"Adding {bin_dir} to PATH in .profile")
    with open(Path.home() / ".profile", "a") as f:
        f.write(f"export PATH=$PATH:{bin_dir}\n")

    os.system("source ~/.profile")
    print("Done!")


def main():

    if len(sys.argv) > 1 and sys.argv[1] == "install":
        install()
        return

    asyncio.run(init_db())
    cli()


if __name__ == "__main__":
    main()



import json
import os
from dataclasses import dataclass
from enum import Enum

from assistants.lib.constants import CLAUDE_CLI_MAX_TOKENS


class CustomKeyNames(str, Enum):
    ASSISTANTS_API_KEY_NAME: str = "ASSISTANTS_API_KEY_NAME"
    ANTHROPIC_API_KEY_NAME: str = "ANTHROPIC_API_KEY_NAME"


@dataclass
class KeyNames:
    ASSISTANTS_API_KEY_NAME: str
    ANTHROPIC_API_KEY_NAME: str


def get_keynames():
    return KeyNames(
        ASSISTANTS_API_KEY_NAME=os.environ.get(
            CustomKeyNames.ASSISTANTS_API_KEY_NAME, "OPENAI_API_KEY"
        ),
        ANTHROPIC_API_KEY_NAME=os.environ.get(
            CustomKeyNames.ANTHROPIC_API_KEY_NAME, "ANTHROPIC_API_KEY"
        ),
    )


@dataclass
class Config:
    OPENAI_API_KEY: str
    ANTHROPIC_API_KEY: str
    DEFAULT_MODEL: str
    CODE_MODEL: str
    IMAGE_MODEL: str
    ASSISTANT_INSTRUCTIONS: str
    ASSISTANT_NAME: str
    TELEGRAM_BOT_TOKEN: str
    CLAUDE_MAX_TOKENS: str
    OPEN_IMAGES_IN_BROWSER: bool


def get_config() -> Config:
    custom_keys = get_keynames()
    return Config(
        OPENAI_API_KEY=os.environ.get(custom_keys.ASSISTANTS_API_KEY_NAME, None),
        ANTHROPIC_API_KEY=os.environ.get(custom_keys.ANTHROPIC_API_KEY_NAME, None),
        DEFAULT_MODEL=os.environ.get("DEFAULT_MODEL", "gpt-4o-mini"),
        CODE_MODEL=os.environ.get("CODE_MODEL", "o1-mini"),
        IMAGE_MODEL=os.environ.get("IMAGE_MODEL", "dall-e-3"),
        ASSISTANT_INSTRUCTIONS=os.environ.get(
            "ASSISTANT_INSTRUCTIONS", "You are a helpful assistant."
        ),
        ASSISTANT_NAME=os.environ.get("ASSISTANT_NAME", "DefaultAssistant"),
        TELEGRAM_BOT_TOKEN=os.environ.get("TG_BOT_TOKEN", None),
        CLAUDE_MAX_TOKENS=os.environ.get("CLAUDE_MAX_TOKENS", CLAUDE_CLI_MAX_TOKENS),
        OPEN_IMAGES_IN_BROWSER=bool(
            json.loads(os.environ.get("OPEN_IMAGES_IN_BROWSER", "true"))
        ),
    )


environment = get_config()



from setuptools import find_packages, setup

from assistants import version

setup(
    name="assistants-framework",
    version=version.__VERSION__,
    author="Michael Jarvis",
    author_email="nihilok@jarv.dev",
    description="OpenAI Assistants Wrapper, with CLI and Telegram Bot",
    long_description=open("README.md").read() + "\n\n" + open("CHANGELOG.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/nihilok/assistants",
    packages=find_packages(exclude=["assistants.tests*"]),
    install_requires=[
        "openai==1.65.4",
        "aiosqlite==0.20.0",
        "loguru==0.7.3",
        "pyperclip==1.9.0",
        "prompt-toolkit==3.0.48",
        "pygments==2.18.0",
        "anthropic==0.49.0",
        "aiofiles==24.1.0",
        "aiohttp==3.11.11",
    ],
    extras_require={
        "telegram": [
            # Dependencies for the Telegram bot integration
            "python-telegram-bot==21.10",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            "ai-cli=assistants.main:main",
            "ai-tg-bot=assistants.main_tg:main",
            "claude=assistants.claude:main",
        ],
    },
    python_requires=">=3.10",
    keywords="openai gpt3 gpt3.5 gpt4 o1 chatgpt chatbot assistant assistants claude anthropic cli telegram llm bot ui tui coding-assistant coding programming",
)



import curses
import math
from dataclasses import dataclass
from typing import Any


@dataclass
class TerminalSelectorOption:
    label: str
    value: Any


class TerminalSelector:
    def __init__(
        self, items: list[TerminalSelectorOption], title: str = "Please select an item"
    ):
        self.items = items
        self.title = title
        self.current_position = 0
        self.window_start = 0

    def draw_menu(self, stdscr):
        height, width = stdscr.getmaxyx()
        # Reserve one line for title, one for instructions
        max_display_items = height - 3

        # Clear screen
        stdscr.clear()

        # Display title
        if width > 0:
            stdscr.addstr(0, 0, self.title[:width], curses.A_BOLD)

        # Display instructions
        if width > 0:
            stdscr.addstr(
                height - 1,
                0,
                self.truncate(
                    "Use ↑/↓ arrows to navigate, Enter to select, q to quit", width
                ),
                curses.A_DIM,
            )

        # Adjust window if current position is out of view
        if self.current_position >= self.window_start + max_display_items:
            self.window_start = self.current_position - max_display_items + 1
        elif self.current_position < self.window_start:
            self.window_start = self.current_position

        # Display items
        for idx in range(min(max_display_items, len(self.items) - self.window_start)):
            item_idx = idx + self.window_start
            item = self.items[item_idx].label

            # Truncate item if it's too long
            item = self.truncate(item, width)

            # Highlight current selection
            if item_idx == self.current_position:
                if len(item) < width - 4:
                    item += " " * (width - 4 - len(item))
                stdscr.attron(curses.A_REVERSE)
                stdscr.addstr(idx + 1, 0, f" {item}"[:width])
                stdscr.attroff(curses.A_REVERSE)
            else:
                stdscr.addstr(idx + 1, 0, f" {item} "[:width])

        # Display scrollbar if needed
        if len(self.items) > max_display_items and width > 0:
            scrollbar_height = math.ceil(
                (max_display_items / len(self.items)) * max_display_items
            )
            scrollbar_pos = math.floor(
                (self.window_start / len(self.items)) * max_display_items
            )
            for i in range(max_display_items):
                if scrollbar_pos <= i < scrollbar_pos + scrollbar_height:
                    stdscr.addstr(i + 1, width - 1, "█")
                else:
                    stdscr.addstr(i + 1, width - 1, "│")

        stdscr.refresh()

    @staticmethod
    def truncate(item, width):
        if len(item) > width - 4:
            item = item[: width - 7] + "..."
        return item

    def run(self):
        def _inner(stdscr):
            # Setup
            curses.curs_set(0)  # Hide cursor
            stdscr.clear()

            while True:
                self.draw_menu(stdscr)
                key = stdscr.getch()

                if key == curses.KEY_UP and self.current_position > 0:
                    self.current_position -= 1
                elif (
                    key == curses.KEY_DOWN
                    and self.current_position < len(self.items) - 1
                ):
                    self.current_position += 1
                elif key in [curses.KEY_ENTER, ord("\n"), ord("\r")]:
                    return self.items[self.current_position].value
                elif key == ord("q"):
                    return None

        return curses.wrapper(_inner)


# Example usage
if __name__ == "__main__":
    # Sample list of items
    _items = [
        TerminalSelectorOption(label=f"Item {i}", value=i) for i in range(1, 101)
    ]  # Create 100 items

    selector = TerminalSelector(_items, "Select an item from the list")
    selected_item = selector.run()

    # Print result after exiting curses window
    if selected_item:
        print(f"You selected: {selected_item}")
    else:
        print("Selection cancelled")



import argparse

from assistants import version
from assistants.config import environment


def get_args():
    parser = argparse.ArgumentParser(
        description=f"CLI for assistants-framework v{version.__VERSION__}",
        epilog=f"""
OPENAI_API_KEY environment variable {'must be set to use the OpenAI API (not set)' if not environment.OPENAI_API_KEY else 'is set.'}
{"ANTHROPIC_API_KEY environment variable must be set to use the Anthropic API (not set)" if environment.CODE_MODEL.startswith("claude-") and not environment.ANTHROPIC_API_KEY else "ANTHROPIC_API_KEY environment variable is set." if environment.ANTHROPIC_API_KEY and environment.CODE_MODEL.startswith('claude-') else ""}
""",
    )
    parser.add_argument(
        "-e",
        "--editor",
        action="store_true",
        help="open the default editor to compose a prompt.",
    )
    parser.add_argument(
        "-f",
        "--input-file",
        metavar="INPUT_FILE",
        type=str,
        help="read the initial prompt from a file (e.g., 'input.txt').",
    )
    parser.add_argument(
        "-i",
        "--instructions",
        metavar="INSTRUCTIONS_FILE",
        type=str,
        help="read the initial instructions (system message) from a specified file; "
        "if this file is not provided, environment variable `ASSISTANT_INSTRUCTIONS` "
        "will be used (or a default of 'You are a helpful assistant').",
    )
    parser.add_argument(
        "-t",
        "--continue-thread",
        action="store_true",
        help="continue previous thread.",
    )
    parser.add_argument(
        "-C",
        "--code",
        action="store_true",
        help="use specialised reasoning/code model. WARNING: This model may be slower "
        "and more expensive to use (use the CODE_MODEL environment variable to change "
        "the model used. Defaults to 'o1-mini').",
    )
    parser.add_argument(
        "prompt",
        nargs="*",
        help="positional arguments concatenate into a single prompt. E.g. `ai-cli "
        "Is this a single prompt\\?` (question mark escaped)\n"
        "...will be passed to the program as a single string (without the backslash). You "
        "can also use quotes to pass a single argument with spaces and special characters. "
        "See the -e and -f options for more advanced prompt options.",
    )
    args = parser.parse_args()
    return args
